# -*- coding:utf-8 -*-

import lbaas_pb2
import google.protobuf.json_format as jf
import traceback
import base64
import re

from oslo_log import log as logging

from netaddr.strategy import ipv6

LOG = logging.getLogger(__name__)


def build_wrapper(f):
    def caller(*args, **kargs):
        try:
            return f(*args, **kargs)
        except Exception as e:
            LOG.error("[build service] %s error: %s" % (f, traceback.format_exc()))
            raise e
    return caller


class UuidBuilder(object):
    def __init__(self, uuid):
        self.uuid = uuid
        self.clean = True

    @build_wrapper
    def build(self):
        uuid = lbaas_pb2.service()
        jf.ParseDict({"uuid": self.uuid, "clean": self.clean}, uuid)
        return base64.b64encode(uuid.SerializeToString())


class ServiceBuilder(object):
    def __init__(self, driver, service):
        self.driver = driver
        self.service = service

        self.loadbalancer = {}
        self.listeners = {}
        self.pools = {}
        self.members = {}
        self.healthmonitors = {}
        self.networks = {}
        self.subnets = {}

        self.need_parse = True

    @build_wrapper
    def build(self, need_parse=False):
        if need_parse or self.need_parse:
            self._parse()

        maps = self._build_service()
        # LOG.debug("[build service] maps result: %s" % maps)

        if maps is None:
            LOG.debug("[build service] maps is empty")
            return None

        LOG.debug("[build service]: %s" % maps)
        service = lbaas_pb2.service()
        jf.ParseDict(maps, service)
        return base64.b64encode(service.SerializeToString())

    def _build_netaddr(self, addr):
        if not addr:
            return addr

        if ipv6.valid_str(addr):
            return ipv6.int_to_str(ipv6.str_to_int(addr), ipv6.ipv6_full)
        else:
            return addr

    def _build_net(self, net):
        (n, m) = net.split('/')
        return self._build_netaddr(n) + '/' + m

    @build_wrapper
    def _build_service(self):
        if len(self.loadbalancer) != 1:
            LOG.error("[build service] build service failed, no loadbalancer here")
            return None
        config = self.service["loadbalancer"]
        loadbalancer = self._build_loadbalancer(config)
        return {"loadbalancer": loadbalancer}

    @build_wrapper
    def _build_loadbalancer(self, config):
        lb = dict()
        lb["name"] = config.get("id")
        if config.get("name", "") != "":
            lb["prefix"] = config["name"] + "-"
        lb["vip"] = self._build_netaddr(config.get("vip_address"))
        lb["tenant"] = config.get("tenant_id")
        lb["mac"] = config.get('vip_port', {}).get('mac_address')

        subnet = self.subnets[config.get("vip_subnet_id")]
        lb["mask"] = subnet['cidr'].split('/')[1]
        lb["gateway"] = self._build_netaddr(subnet['gateway_ip'])

        lb['routes'] = []
        if subnet['gateway_ip']:
            if ipv6.valid_str(subnet['gateway_ip']):
                lb['routes'] = [{'net': '0:0:0:0:0:0:0:0/0', 'gw': lb['gateway']}]
            else:
                lb['routes'] = [{'net': '0.0.0.0/0', 'gw': lb['gateway']}]

        for route in subnet['host_routes']:
            lb['routes'].append({'net': self._build_net(route['destination']), 
                'gw': self._build_netaddr(route['nexthop'])})

        if config.get("network_id") is not None:
            lb["network"] = self._build_network(self.networks[config.get("network_id")])

        lb["listeners"] = []
        for m in config.get("listeners", []):
            listener = self._build_listener(self.listeners[m["id"]])
            if listener:
                lb["listeners"].append(listener)

        lb["ports"] = []
        for p in self.driver.get_snat_ports(self.service):
            port = self._build_port(config, p)
            if port:
                lb["ports"].append(port)

        return lb

    @build_wrapper
    def _build_port(self, lb, config):
        port = {}

        network = self.networks[config["network_id"]]
        subnet = self.subnets[config['fixed_ips'][0]['subnet_id']]

        network = self.driver.get_snat_network(config, network)
        port["name"] = lb["id"] + "-" + subnet["id"][:8]
        port["network"] = self._build_network(network)
        port["ip"] = self._build_netaddr(config['fixed_ips'][0]['ip_address'])
        port["mask"] = subnet['cidr'].split('/')[1]

        port["mac"] = config['mac_address']

        return port

    @build_wrapper
    def _build_network(self, config):
        if not config:
            return None

        network = dict()
        if config.get("provider:network_type") == "vlan":
            network["type"] = "vlan"
            network["vlan"] = self._build_vlan(config)
        elif config.get("provider:network_type") == "vxlan":
            network["type"] = "vxlan"
            network["vxlan"] = self._build_vxlan(config)
        else:
            LOG.error("[build service] unknown network type: %s" % config.get("provider:network_type", None))

        return network

    @build_wrapper
    def _build_vlan(self, config):
        if not config:
            return None

        vlan = dict()
        vlan["name"] = config.get("id")
        interface, tagged = self.driver.get_vlan_tuple(config)
        vlan["interface"] = interface
        vlan["vlanid"] = int(config.get('provider:segmentation_id', 0)) if tagged else 0

        return vlan

    @build_wrapper
    def _build_vxlan(self, config):
        if not config:
            return None

        vxlan = dict()
        vxlan["name"] = config.get("id")
        vxlan["vxlanid"] = config.get("provider:segmentation_id", 0)
        vxlan["domain"] = 0
        return vxlan

    @build_wrapper
    def _build_rule(self, config):
        rule = dict()
        rule['name'] = config.get('id')

        type = config['type']
        compare_type = config['compare_type']
        header = config.get('key')
        value = config['value']

        if config['type'] == "PATH":
            type = "URI"
            compare_type = config['compare_type']
        elif config['type'] == 'FILE_TYPE':
            type = "URI"
            compare_type = "ENDS_WITH"
        elif config['type'] == 'HOST_NAME':
            type = "HEADER"
            header = "Host"
        elif config['type'] == 'COOKIE':
            type = "HEADER"
            header = "Cookie"
            K = re.escape(config['key'])
            V = re.escape((config['value']))
            if compare_type == "STARTS_WITH":
                value = "(.*;|^)\s*%s\s*=\s*%s.*" % (K, V)
            elif compare_type == "ENDS_WITH":
                value = "(.*;|^)\s*%s\s*=\s*[^;]*%s(;|$)" % (K, V)
            elif compare_type == "EQUAL_TO":
                value = "(.*;|^)\s*%s\s*=\s*%s(;|$)" % (K, V)
            elif compare_type == "CONTAINS":
                value = "(.*;|^)\s*%s\s*=[^;]*%s" % (K, V)
            else:
                value = "(.*;|^)\s*%s\s*=\s*%s" % (r['key'], r['value'])
            compare_type = 'REGEX'

        V = re.escape(config['value'])
        if compare_type == "STARTS_WITH":
            compare_type = "STR_MATCH_TYPE_REGEX"
            value = "^%s.*" % V
        elif compare_type == "ENDS_WITH":
            compare_type = "STR_MATCH_TYPE_REGEX"
            value = ".*%s$" % V
        elif compare_type == "REGEX":
            compare_type = "STR_MATCH_TYPE_REGEX"
        elif compare_type == "EQUAL_TO":
            compare_type = "STR_MATCH_TYPE_EQUAL"
        elif compare_type == "CONTAINS":
            compare_type = "STR_MATCH_TYPE_CONTAIN"
        else:
            compare_type = "STR_MATCH_TYPE_NONE"

        rule['type'] = type
        if header is not None:
            rule['key'] = header
        rule['compare_type'] = compare_type
        rule['value'] = value
        rule['invert'] = config['invert']

        return rule

    @build_wrapper
    def _build_policy(self, config):
        policy = dict()
        policy['name'] = config.get('id')
        if config.get('name', '') != '':
            policy['prefix'] = config['name'] + "-"
        policy['enable'] = config.get('admin_state_up', False)
        if config['action'] == "REJECT":
            policy['action'] = "RULE_ACTION_DROP"
        elif config['action'] == "REDIRECT_TO_POOL":
            policy['action'] = "RULE_ACTION_SCHED"
            policy['pool'] = config['redirect_pool_id']
        elif config['action'] == 'REDIRECT_TO_URL':
            policy['action'] = "RULE_ACTION_REDIR_URL"
            policy['url'] = config['redirect_url']
        else:
            assert 0

        policy['rules'] = []
        for i in config.get("l7rules", []):
            rule = self.l7rules[i]
            policy['rules'].append(self._build_rule(rule))

        return policy


    @build_wrapper
    def _build_listener(self, config):
        listener = dict()
        listener["name"] = config.get("id")
        if config.get('name', '') != '':
            listener['prefix'] = config['name'] + "-"
        if config.get('protocol') == 'HTTP':
            listener["srvtype"] = 'SRV_HTTP'
            listener["mode"] = 'VS_MODE_L7'
        elif config.get('protocol') == 'HTTPS':
            listener["srvtype"] = 'SRV_HTTPS'
            listener["mode"] = 'VS_MODE_L7'
        else:
            listener["srvtype"] = 'SRV_TCP'
            listener["mode"] = 'VS_MODE_L4'
        #listener["srvtype"] = 'SRV_HTTP' if config.get('protocol') == 'HTTP' else 'SRV_TCP'
        #listener["mode"] = 'VS_MODE_L7' if config.get('protocol') == 'HTTP' else 'VS_MODE_L4'
        listener["enable"] = config.get('admin_state_up', False)
        listener["port"] = int(config.get('protocol_port'))
        listener["limit"] = config.get("connection_limit", -1)
        listener["x_forward_enable"] = self.driver.conf.x_forwarded_for_enable

        if config.get("default_pool_id") is not None:
            pool = self.pools[config.get("default_pool_id")]
            listener["pool"] = self._build_pool(pool)

        listener["policies"] = []
        for i in config.get("l7policies", []):
            policy = self.l7policies[i]
            listener['policies'].append(self._build_policy(policy))

        return listener

    @build_wrapper
    def _build_pool(self, config):
        if not config:
            return None

        pool = dict()
        pool["name"] = config.get("id")
        if config.get('name', '') != '':
            pool['prefix'] = config['name'] + "-"

        algorithm = config.get("lb_algorithm")
        if algorithm == "ROUND_ROBIN":
            pool["method"] = "NODE_LB_WRR"
        elif algorithm == "LEAST_CONNECTIONS":
            pool["method"] = "NODE_LB_LEAST_CONN"
        else:
            pool["method"] = "NODE_LB_HASH"

        pool["persistence"] = self._build_persistence(config.get("session_persistence"), config.get("protocol"))

        pool["members"] = []
        for m in config.get("members", []):
            pool["members"].append(self._build_member(self.members[m]))

        if config.get("healthmonitor_id") is not None:
            pool["healthmonitor"] = self._build_healthmonitor(self.healthmonitors[config.get("healthmonitor_id")])

        return pool

    @build_wrapper
    def _build_persistence(self, config, protocol):
        if not config:
            return None
        if protocol != "HTTP" and config.get("type") != "SOURCE_IP":
            LOG.error("[build service] invalid persistence, protocol %s and type %s" % (protocol, config.get("type")))
            return None
        persistence = {}
        if config.get("type") == "SOURCE_IP":
            persistence["type"] = "PERSIST_SOURCE_IP"
        elif config.get("type") == "HTTP_COOKIE":
            persistence["type"] = "PERSIST_COOKIE"
            persistence["cookie_type"] = "PERSIST_COOKIE_INSERT"
            persistence["cookie_name"] = "openstack_cookie_insert"
        else:
            persistence["type"] = "PERSIST_COOKIE"
            persistence["cookie_type"] = "PERSIST_COOKIE_STUDY"
            persistence["cookie_name"] = config.get('cookie_name')
        return persistence

    @build_wrapper
    def _build_member(self, config):
        if not config:
            return None

        member = dict()
        member["address"] = self._build_netaddr(config.get("address"))
        member["port"] = config.get("protocol_port")
        member["weight"] = config.get("weight")
        member["enable"] = config.get("admin_state_up", True)
        if member["weight"] == 0:
            member["enable"] = False
            member["weight"] = 1

        return member

    @build_wrapper
    def _build_healthmonitor(self, config):
        if not config:
            return None

        monitor = dict()
        monitor["name"] = config.get("id")
        if config.get('name', '') != '':
            monitor['prefix'] = config['name'] + "-"

        if config.get("type") == "TCP":
            monitor["type"] = "MONITOR_CONNECT_TCP"
        elif config.get("type") == "PING":
            monitor["type"] = "MONITOR_ICMPV4"
        elif config.get("type") == "HTTP":
            monitor["type"] = "MONITOR_CONNECT_HTTP"
        elif config.get("type") == "HTTPS":
            monitor["type"] = "MONITOR_CONNECT_HTTPS"
        else:
            monitor["type"] = "MONITOR_ICMPV4"

        if config.get("type") == "HTTP" or config.get("type") == "HTTPS":
            codes = config.get("expected_codes", "200")
            if codes.find('-') != -1:
                (s, e) = codes.split('-')
                t = []
                for i in range(int(s), int(e) + 1):
                    t.append(str(i))
                monitor['codes'] = ';'.join(t)
            else:
                monitor['codes'] = codes.replace(',', ';')
            monitor["url"] = "%s %s HTTP/1.0" % (config.get("http_method", "GET"), config.get("url_path", "/"))

        monitor["retry"] = int(config.get("max_retries"))
        if monitor["retry"] > 255:
            monitor["retry"] = 255
        monitor["timeout"] = int(config.get("timeout"))
        if monitor["timeout"] > 255:
            monitor["timeout"] = 255
        monitor["delay"] = int(config.get("delay"))
        if monitor["delay"] > 255:
            monitor["delay"] = 255

        return monitor

    @build_wrapper
    def _parse(self):
        # LOG.debug("[build service] [proto parse] %s" % self.service)
        service = self.service
        if "loadbalancer" in service:
            self._parse_element("loadbalancer")
        if "listeners" in service:
            #self._parse_elements("listeners")
            self.listeners = service["listeners"]
        if "pools" in service:
            self.pools = service["pools"]
        if "members" in service:
            self.members = service["members"]
        if "healthmonitors" in service:
            self.healthmonitors = service["healthmonitors"]
        if "networks" in service:
            self.networks = service["networks"]
        if "subnets" in service:
            self.subnets = service["subnets"]
        if "l7policies" in service:
            self.l7policies = service['l7policies']
        if 'l7rules' in service:
            self.l7rules = service['l7rules']
        self.need_parse = False

    @build_wrapper
    def _parse_element(self, name):
        getattr(self, name)
        e = self.service[name]
        if "id" in e:
            setattr(self, name, {e["id"]: e})

    @build_wrapper
    def _parse_elements(self, name):
        getattr(self, name)
        maps = {}
        elements = self.service[name].values()
        for e in elements:
            if "id" in e:
                maps[e["id"]] = e
        setattr(self, name, maps)
