# coding=utf-8
# Copyright 2016 F5 Networks Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
try:
    version = 'q'
    from neutron.db import segments_db as sdb
    from neutron.plugins.ml2 import db
except:
    version = 'm'
    from neutron.plugins.ml2 import db
    sdb = db
from neutron.plugins.ml2 import models
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class DisconnectedService(object):
    def __init__(self):
        self.supported_encapsulations = ['vlan']

    # Retain this method for future use in case a particular ML2 implementation
    # decouples network_id from physical_network name.  The implementation in
    # neutron db.py requires a network_id.
    def get_network_segments(self, context, host, port_id, network_id):
        if version == 'q':
            return sdb.get_network_segments(context,
                                       network_id, filter_dynamic=None)
        else:
            return sdb.get_network_segments(context.session,
                                       network_id, filter_dynamic=None)

    def get_network_segment(self, context, agent_configuration,
                            host, port_id, network):
        data = None

        network_physical_maps = \
            agent_configuration.get('bridge_mappings', {})

        supported_encapsulations = [
            x.lower() for x in self.supported_encapsulations +
            agent_configuration.get('tunnel_types', [])
        ]

        segments = self.get_network_segments(context,
                                             host,
                                             port_id,
                                             network['id'])

        for segment in segments:
            LOG.debug("Sangfor disconnected service check segment: %s" % segment)
            if ((not segment['physical_network']
                 or segment['physical_network'] in network_physical_maps
                 or 'default' in network_physical_maps)
                    and (segment['network_type'].lower()
                         in supported_encapsulations)):
                data = segment
                break

        if not data:
            LOG.debug('Using default segment for network %s' %
                      (network['id']))

            # neutron is expected to provide this data immediately
            data = {
                'segmentation_id': network['provider:segmentation_id']
            }
            if 'provider:network_type' in network:
                data['network_type'] = network['provider:network_type']
            if 'provider:physical_network' in network:
                data['physical_network'] = network['provider:physical_network']

        return data


class DisconnectedServicehuawei(DisconnectedService):

    def get_network_segments(self, context, host, port_id, network_id):
        """
        1.find segment_id from binding levels table 
        2.return segment info from ml2 segment table
        """
        if version == 'q':
            ctx = context
        else:
            ctx = context.session
        segment_infos = []
        binding_levels = db.get_binding_levels(ctx, port_id, host)
        LOG.debug('port %s binding level info (%s)'%(port_id, binding_levels))
        for binding_level in binding_levels:
            LOG.debug('binding info (%s)'%(binding_level))
            segment_info = sdb.get_segment_by_id(ctx, binding_level.segment_id)
            if segment_info:
                segment_infos.append(segment_info)

        return segment_infos


class DisconnectedServiceTest(DisconnectedService):
    """
    内部测试用，vxlan网络直接转vlan下发
    """
    def get_network_segments(self, context, host, port_id, network_id):
        segments = super(DisconnectedServiceTest, self)\
            .get_network_segments(context, host, port_id, network_id)
        for segment in segments:
            if (segment['network_type'].lower() == 'vxlan'
                    and segment['segmentation_id'] < 4095):
                segment['network_type'] = 'vlan'
        return segments

