# coding=utf-8

import os
import sys
import uuid

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import importutils

from neutron.common import constants as q_const
try:
    from neutron.extensions.portbindings import portbindings
except:
    from neutron.extensions import portbindings
try:
    from neutron_lib import constants as plugin_constants
except:
    from neutron.plugins.common import constants as plugin_constants
from neutron_lbaas.services.loadbalancer import constants as lb_const

from neutron_lbaas.db.loadbalancer import models
from neutron_lbaas.extensions import lbaas_agentschedulerv2

from sangforlbaasdriver.v2.ad import agent_rpc
from sangforlbaasdriver.v2.ad import exceptions as sangfor_exc
from sangforlbaasdriver.v2.ad import plugin_rpc
from sangforlbaasdriver.v2.ad import constants_v2 as constants
from sangforlbaasdriver import __version__

import traceback

LOG = logging.getLogger(__name__)

OPTS = [
    cfg.StrOpt(
        'sangfor_loadbalancer_pool_scheduler_driver_v2',
        default=(
            'sangforlbaasdriver.v2.ad.agent_scheduler.TenantScheduler'
        ),
        help=('Driver to use for scheduling '
              'pool to a default loadbalancer agent')
    ),
    cfg.StrOpt(
        'sangfor_loadbalancer_service_builder_v2',
        default=(
            'sangforlbaasdriver.v2.ad.service_builder.LBaaSv2ServiceBuilder'
        ),
        help=('Default class to use for building a service object.')
    ),
    cfg.StrOpt('sangfor_host_id',
               default=None,
               help=('host id for loadbalancer port')),
    cfg.StrOpt('sangfor_device_owner',
               default='neutron:LOADBALANCERV2',
               help=('device owner for loadbalancer port')),
]

class SangforNoAttachedLoadbalancerException(sangfor_exc.SangforLBaaSv2DriverException):
    """Exception thrown when an LBaaSv2 object has not parent Loadbalancer."""

    message = "Entity has no associated loadalancer"

class SangforDriverV2(object):

    def __init__(self, plugin=None, env=None):

        if not plugin:
            LOG.error("Required LBaaS Driver and Core Driver Missing")
            sys.exit(1)

        self.plugin = plugin
        self.env = env

        self.loadbalancer = LoadBalancerManager(self)
        self.listener = ListenerManager(self)
        self.pool = PoolManager(self)
        self.member = MemberManager(self)
        self.healthmonitor = HealthMonitorManager(self)
        self.l7policy = L7PolicyManager(self)
        self.l7rule = L7RuleManager(self)

        self.conf = self.init_config()
        self.scheduler = importutils.import_object(
            self.conf.sangfor_loadbalancer_pool_scheduler_driver_v2)

        self.service_builder = importutils.import_object(
            self.conf.sangfor_loadbalancer_service_builder_v2, self)

        self.agent_rpc = agent_rpc.LBaaSv2AgentRPC(self)
        self.plugin_rpc = plugin_rpc.LBaaSv2PluginCallbacksRPC(self)

        # add this agent RPC to the neutron agent scheduler
        # mixins agent_notifiers dictionary for it's env
        try:
            agent_type = plugin_constants.AGENT_TYPE_LOADBALANCER
        except:
            agent_type = q_const.AGENT_TYPE_LOADBALANCER
        self.plugin.agent_notifiers.update(
            {agent_type: self.agent_rpc})

    def init_config(self):
        # 使用自己的配置文件，而不是全局的
        default_config = '/etc/neutron/plugins/sangfor/' \
                         'sangfor-openstack-lbaas.ini'
        config_path = '/etc/neutron/plugins/sangfor/' \
                      '%s-openstack-lbaas.ini' % self.env
        if not os.path.exists(config_path):
            config_path = default_config
        LOG.debug('sangfor plugin evn [%s] use config file: %s' %
                  (self.env, config_path))
        config = cfg.ConfigOpts()
        config(args=['--config-file=%s' % config_path], project=self.env,
               version='sangfor %s' % __version__)
        config.register_opts(OPTS)
        return config

class LoadBalancerManager(object):
    """LoadBalancerManager class handles Neutron LBaaS CRUD."""

    def __init__(self, driver):
        """Initialize Loadbalancer manager."""
        self.driver = driver

    @log_helpers.log_method_call
    def create(self, context, loadbalancer):
        """Create a loadbalancer."""
        # 重新加载才能得到provider信息
        loadbalancer = self.driver.plugin.\
            db.get_loadbalancer(context, loadbalancer.id)
        driver = self.driver
        try:
            agent = driver.scheduler.schedule(
                driver.plugin,
                context,
                loadbalancer.id,
                driver.env
            )
            
            # Update the port for the VIP to show ownership by this driver
            try:
                status = q_const.PORT_STATUS_ACTIVE
            except:
                status = plugin_constants.PORT_STATUS_ACTIVE
            port_data = {
                'admin_state_up': True,
                'device_id': loadbalancer.id,
                #'device_owner': 'neutron:LOADBALANCERV2',
                'device_owner': self.driver.conf.sangfor_device_owner,
                'status': status
            }
            """
            看移动的文档，sdn是在create_port_postcommit的时候生成绑定信息的，
            此时plugin是无法获取到provider信息的，因此移动的实现只支持一个
            lbaasv2 plugin，因为多个无法相互区别
            根据华为数广测试通过的情况来看，sdn是在update_port_postcommit
            的时候生成绑定信息的，因此可以在service build之前更新host_id，
            我们可以根据provider生成host信息，同一个集群，异地provider可以
            使用sangfor.nanshan和sangfor.futian来表示同一个host
            """
            if self.driver.conf.sangfor_host_id:
                host_id = self.driver.conf.sangfor_host_id
            else:
                host_id = loadbalancer.provider.provider_name.split('.')[0]
            port_data[portbindings.HOST_ID] = host_id

            driver.plugin.db._core_plugin.update_port(
                context,
                loadbalancer.vip_port_id,
                {'port': port_data}
            )
            service = driver.service_builder.build(context,
                                                   loadbalancer,
                                                   agent=agent)

            driver.agent_rpc.create_loadbalancer(
                context,
                loadbalancer.to_api_dict(),
                service,
                agent['host']
            )

        except (lbaas_agentschedulerv2.NoEligibleLbaasAgent,
                lbaas_agentschedulerv2.NoActiveLbaasAgent,
                sangfor_exc.SangforMismatchedTenants) as e:
            LOG.error("Exception: loadbalancer create: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                loadbalancer.id,
                plugin_constants.ERROR)
        except Exception as e:
            LOG.error("Exception: loadbalancer create: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                loadbalancer.id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def update(self, context, old_loadbalancer, loadbalancer):
        """Update a loadbalancer."""
        driver = self.driver
        try:
            agent = driver.scheduler.schedule(
                driver.plugin,
                context,
                loadbalancer.id,
                driver.env
            )
            service = driver.service_builder.build(context, loadbalancer)
            driver.agent_rpc.update_loadbalancer(
                context,
                old_loadbalancer.to_api_dict(),
                loadbalancer.to_api_dict(),
                service,
                agent['host']
            )
        except (lbaas_agentschedulerv2.NoEligibleLbaasAgent,
                lbaas_agentschedulerv2.NoActiveLbaasAgent) as e:
            LOG.error("Exception: loadbalancer update: %s" % traceback.format_exc())
            driver._handle_driver_error(context,
                                        models.LoadBalancer,
                                        loadbalancer.id,
                                        plugin_constants.ERROR)
        except Exception as e:
            LOG.error("Exception: loadbalancer update: %s" % traceback.format_exc())
            driver._handle_driver_error(context,
                                        models.LoadBalancer,
                                        loadbalancer.id,
                                        plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def delete(self, context, loadbalancer):
        """Delete a loadbalancer."""
        driver = self.driver
        try:
            agent = driver.scheduler.schedule(
                driver.plugin,
                context,
                loadbalancer.id,
                driver.env
            )
            service = driver.service_builder.build(context, loadbalancer, agent)
            driver.agent_rpc.delete_loadbalancer(
                context,
                loadbalancer.to_api_dict(),
                service,
                agent['host']
            )

        except (lbaas_agentschedulerv2.NoEligibleLbaasAgent,
                lbaas_agentschedulerv2.NoActiveLbaasAgent) as e:
            LOG.error("Exception: loadbalancer delete: %s" % traceback.format_exc())
            driver.plugin.db.delete_loadbalancer(context, loadbalancer.id)
        except Exception as e:
            LOG.error("Exception: loadbalancer delete: %s" % traceback.format_exc())
            driver.plugin.db.delete_loadbalancer(context, loadbalancer.id)
            raise e

    @log_helpers.log_method_call
    def refresh(self, context, loadbalancer):
        """Refresh a loadbalancer."""
        pass

    @log_helpers.log_method_call
    def stats(self, context, loadbalancer):
        """Get loadbalancer statistics."""
        pass

class ListenerManager(object):
    """ListenerManager class handles Neutron LBaaS listener CRUD."""

    def __init__(self, driver):
        """Initialize a ListenerManager object."""
        self.driver = driver

    @log_helpers.log_method_call
    def create(self, context, listener):
        """Create a listener."""
        driver = self.driver
        try:
            if listener.attached_to_loadbalancer():
                loadbalancer = listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context,
                                                       loadbalancer,
                                                       agent=agent)

                driver.agent_rpc.create_listener(
                    context,
                    listener.to_dict(loadbalancer=False, default_pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: listener create: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                listener.loadbalancer.id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def update(self, context, old_listener, listener):
        """Update a listener."""
        driver = self.driver
        try:
            if listener.attached_to_loadbalancer():
                loadbalancer = listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.update_listener(
                    context,
                    old_listener.to_dict(loadbalancer=False,
                                         default_pool=False),
                    listener.to_dict(loadbalancer=False, default_pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: listener update: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                listener.loadbalancer.id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def delete(self, context, listener):
        """Delete a listener."""
        driver = self.driver
        try:
            if listener.attached_to_loadbalancer():
                loadbalancer = listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                lb = service['loadbalancer']
                if lb['provisioning_status'] == plugin_constants.PENDING_UPDATE:
                    self.driver.plugin_rpc.update_loadbalancer_status(
                        context, lb['id'], plugin_constants.ACTIVE, lb_const.ONLINE
                    )
                    lb['provisioning_status'] = plugin_constants.ACTIVE
                    LOG.debug("reset ACTIVE for loadbalancer")
                driver.agent_rpc.delete_listener(
                    context,
                    listener.to_dict(loadbalancer=False, default_pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: listener delete: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                listener.loadbalancer.id,
                plugin_constants.ERROR)
            raise e

def pool_get_loadbalancer(pool):
    try:
        lb = pool.loadbalancer
    except:
        lb = pool.listener.loadbalancer

    return lb

class PoolManager(object):
    """PoolManager class handles Neutron LBaaS pool CRUD."""

    def __init__(self, driver):
        """Initialize a PoolManager object."""
        self.driver = driver

    def _get_pool_dict(self, pool):
        pool_dict = pool.to_api_dict()
        pool_dict['provisioning_status'] = pool.provisioning_status
        pool_dict['operating_status'] = pool.operating_status
        return pool_dict

    @log_helpers.log_method_call
    def create(self, context, pool):
        """Create a pool."""
        driver = self.driver
        try:
            if pool.attached_to_loadbalancer():
                loadbalancer = pool_get_loadbalancer(pool)
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context,
                                                       loadbalancer,
                                                       agent=agent)
                driver.agent_rpc.create_pool(
                    context,
                    self._get_pool_dict(pool),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: pool create: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(pool).id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def update(self, context, old_pool, pool):
        """Update a pool."""
        driver = self.driver
        try:
            if pool.attached_to_loadbalancer():
                loadbalancer = pool.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.update_pool(
                    context,
                    self._get_pool_dict(old_pool),
                    self._get_pool_dict(pool),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: pool update: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(pool).id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def delete(self, context, pool):
        """Delete a pool."""
        driver = self.driver
        try:
            if pool.attached_to_loadbalancer():
                loadbalancer = pool.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.delete_pool(
                    context,
                    self._get_pool_dict(pool),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: pool delete: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(pool).id,
                plugin_constants.ERROR)
            raise e


class MemberManager(object):
    """MemberManager class handles Neutron LBaaS pool member CRUD."""

    def __init__(self, driver):
        """Initialize a MemberManager object."""
        self.driver = driver

    @log_helpers.log_method_call
    def create(self, context, member):
        """Create a member."""
        driver = self.driver
        try:
            if member.attached_to_loadbalancer():
                loadbalancer = pool_get_loadbalancer(member.pool)
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context,
                                                       loadbalancer,
                                                       agent=agent)
                lb = service['loadbalancer']
                if lb['provisioning_status'] == plugin_constants.PENDING_UPDATE:
                    self.driver.plugin_rpc.update_loadbalancer_status(
                        context, lb['id'], plugin_constants.ACTIVE, lb_const.ONLINE
                    )
                    lb['provisioning_status'] = plugin_constants.ACTIVE
                    LOG.debug("reset ACTIVE for loadbalancer")
                driver.agent_rpc.create_member(
                    context,
                    member.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: member create: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(member.pool).id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def update(self, context, old_member, member):
        """Update a member."""
        try:
            driver = self.driver
            if member.attached_to_loadbalancer():
                loadbalancer = member.pool.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.update_member(
                    context,
                    old_member.to_dict(pool=False),
                    member.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: member update: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(member.pool).id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def delete(self, context, member):
        """Delete a member."""
        driver = self.driver
        try:
            if member.attached_to_loadbalancer():
                loadbalancer = member.pool.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                lb = service['loadbalancer']
                if lb['provisioning_status'] == plugin_constants.PENDING_UPDATE:
                    self.driver.plugin_rpc.update_loadbalancer_status(
                        context, lb['id'], plugin_constants.ACTIVE, lb_const.ONLINE
                    )
                    lb['provisioning_status'] = plugin_constants.ACTIVE
                    LOG.debug("reset ACTIVE for loadbalancer")

                driver.agent_rpc.delete_member(
                    context,
                    member.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: member delete: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(member.pool).id,
                plugin_constants.ERROR)
            raise e

class HealthMonitorManager(object):
    """HealthMonitorManager class handles Neutron LBaaS monitor CRUD."""

    def __init__(self, driver):
        """Initialize a HealthMonitorManager object."""
        self.driver = driver

    @log_helpers.log_method_call
    def create(self, context, health_monitor):
        """Create a health monitor."""
        driver = self.driver
        try:
            if health_monitor.attached_to_loadbalancer():
                loadbalancer = pool_get_loadbalancer(health_monitor.pool)
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context,
                                                       loadbalancer,
                                                       agent=agent)
                driver.agent_rpc.create_health_monitor(
                    context,
                    health_monitor.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: health monitor create: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(health_monitor.pool).id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def update(self, context, old_health_monitor, health_monitor):
        """Update a health monitor."""
        driver = self.driver
        try:
            if health_monitor.attached_to_loadbalancer():
                loadbalancer = health_monitor.pool.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.update_health_monitor(
                    context,
                    old_health_monitor.to_dict(pool=False),
                    health_monitor.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: health monitor update: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(health_monitor.pool).id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def delete(self, context, health_monitor):
        """Delete a health monitor."""
        driver = self.driver
        try:
            if health_monitor.attached_to_loadbalancer():
                loadbalancer = health_monitor.pool.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.delete_health_monitor(
                    context,
                    health_monitor.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: health monitor: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                pool_get_loadbalancer(health_monitor.pool).id,
                plugin_constants.ERROR)
            raise e

class L7PolicyManager(object):
    """L7PolicyManager class handles Neutron L7Policy CRUD"""
    def __init__(self, driver):
        """Initialize a L7PolicyManager object."""
        self.driver = driver

    @log_helpers.log_method_call
    def create(self, context, policy):
        """Create a l7policy."""
        driver = self.driver
        try:
            if policy.attached_to_loadbalancer():
                loadbalancer = policy.listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context,
                                                       loadbalancer,
                                                       agent=agent)
                driver.agent_rpc.create_l7policy(
                    context,
                    policy.to_dict(),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: l7policy create: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                loadbalancer.id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def update(self, context, old_policy, policy):
        """Update a l7policy."""
        driver = self.driver
        try:
            if policy.attached_to_loadbalancer():
                loadbalancer = policy.listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.update_l7policy(
                    context,
                    old_policy.to_dict(pool=False),
                    policy.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: l7policy update: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                loadbalancer.id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def delete(self, context, policy):
        """Delete a l7policy."""
        driver = self.driver
        try:
            if policy.attached_to_loadbalancer():
                loadbalancer = policy.listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.delete_l7policy(
                    context,
                    policy.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: l7policy: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                loadbalancer.id,
                plugin_constants.ERROR)
            raise e

class L7RuleManager(object):
    """L7RuleManager class handles Neutron L7Rule CRUD"""
    def __init__(self, driver):
        """Initialize a L7RuleManager object."""
        self.driver = driver

    @log_helpers.log_method_call
    def create(self, context, rule):
        """Create a l7rule."""
        driver = self.driver
        try:
            if rule.attached_to_loadbalancer():
                loadbalancer = rule.policy.listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context,
                                                       loadbalancer,
                                                       agent=agent)
                driver.agent_rpc.create_l7rule(
                    context,
                    rule.to_dict(),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: l7rule create: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                loadbalancer.id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def update(self, context, old_rule, rule):
        """Update a l7rule."""
        driver = self.driver
        try:
            if rule.attached_to_loadbalancer():
                loadbalancer = rule.policy.listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.update_l7rule(
                    context,
                    old_rule.to_dict(pool=False),
                    rule.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: l7rule update: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                loadbalancer.id,
                plugin_constants.ERROR)
            raise e

    @log_helpers.log_method_call
    def delete(self, context, rule):
        """Delete a l7rule."""
        driver = self.driver
        try:
            if rule.attached_to_loadbalancer():
                loadbalancer = rule.policy.listener.loadbalancer
                agent = driver.scheduler.schedule(
                    driver.plugin,
                    context,
                    loadbalancer.id,
                    driver.env
                )
                service = driver.service_builder.build(context, loadbalancer)
                driver.agent_rpc.delete_l7rule(
                    context,
                    rule.to_dict(pool=False),
                    service,
                    agent['host']
                )
            else:
                raise SangforNoAttachedLoadbalancerException()

        except Exception as e:
            LOG.error("Exception: l7rule delete: %s" % traceback.format_exc())
            driver.plugin.db.update_status(
                context,
                models.LoadBalancer,
                loadbalancer.id,
                plugin_constants.ERROR)
            raise e

try:
    from neutron_lib.plugins.ml2 import api
except:
    from neutron.plugins.ml2 import driver_api as api

class SangforLbassExtensionDriver(api.MechanismDriver):
    def initialize(self):
        pass

    def create_port_precommit(self, context):
        """
        context._binding:
        <neutron.plugins.ml2.models.PortBinding[object at 7f65fd097910]
        {
            port_id='694e3cb1-d88c-4492-953c-a8144c76fcc7', host='',
            vnic_type='normal', profile='', vif_type='unbound', vif_details=''
        }>
        context._port:
        {"
            status": "DOWN", "binding:host_id": "", "
            allowed_address_pairs": [], "device_owner": "neutron:LOADBALANCER",
            "binding:profile": {}, "
            fixed_ips": [
            {
                "subnet_id": "75ea53df-252e-4b4b-bcf9-a678409d63c4",
                "ip_address": "192.168.117.17"
            }],
            "id": "694e3cb1-d88c-4492-953c-a8144c76fcc7", "
            security_groups": ["386deaf2-cd33-4bb1-b4e5-d700a0410f8f"],
            "device_id": "", "name": "vip-00a6239d-2d87-454c-849a-9aef261994c4",
            "admin_state_up": false,
            "network_id": "52db8248-76de-4585-b6e0-e7999400ee0d",
            "tenant_id": "37898f0a9bb44fa78d6597aea4de5291",
            "binding:vif_details": {}, "binding:vnic_type": "normal",
            "binding:vif_type": "unbound", "mac_address": "fa:16:3e:bf:77:eb"
        }
        :param context:
        :return:
        """
        sangfor_host_id = 'sangfor'
        port = context._port
        binding = context._binding
        if port['device_owner'] == 'neutron:LOADBALANCERV2'\
                and port[portbindings.VIF_TYPE] == portbindings.VIF_TYPE_UNBOUND:
            binding.host = sangfor_host_id
            binding.vif_type = constants.VIF_TYPE

        # 解决生成lan口vif_type未绑定问题
        unplugged_types = (portbindings.VIF_TYPE_BINDING_FAILED,
                           portbindings.VIF_TYPE_UNBOUND)
        if binding.host.find('sangfor') >= 0 \
                and binding.vif_type in unplugged_types:
            binding.vif_type = constants.VIF_TYPE


