# coding=utf-8
import datetime
import json

from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import importutils

from sangforlbaasdriver.v2.ad import constants_v2
from sangforlbaasdriver.v2.ad import exceptions as sangfor_exc
try:
    from neutron.extensions.portbindings import portbindings
except:
    from neutron.extensions import portbindings
from oslo_config import cfg

LOG = logging.getLogger(__name__)

OPTS = [
    cfg.StrOpt(
        'sangfor_loadbalancer_network_builder_v2',
        default=(
            'sangforlbaasdriver.v2.ad.disconnected_service.DisconnectedService'
        ),
        help=(
            'Default class to use for building a network object.'
        )
    )
]

class LBaaSv2ServiceBuilder(object):
    """The class creates a service definition from neutron database.

    A service definition represents all the information required to
    construct a load-balancing service on AD.

    Requests come in to agent as full service definitions, not incremental
    changes. The driver looks up networks, mac entries, segmentation info,
    etc and places all information in a service object (which is a python
    dictionary variable) and passes that to the agent.

    """

    def __init__(self, driver):
        """Get full service definition from loadbalancer id."""
        self.driver = driver

        self.net_cache = {}
        self.subnet_cache = {}
        self.last_cache_update = datetime.datetime.fromtimestamp(0)
        self.driver.conf.register_opts(OPTS)
        # self.disconnected_service = DisconnectedService()
        self.disconnected_service = importutils.import_object(
            self.driver.conf.sangfor_loadbalancer_network_builder_v2)
        self.plugin = self.driver.plugin

    def build(self, context, loadbalancer, agent=None):
        """Get full service definition from loadbalancer ID."""
        # Invalidate cache if it is too old
        if ((datetime.datetime.now() - self.last_cache_update).seconds >
                constants_v2.NET_CACHE_SECONDS):
            self.net_cache = {}
            self.subnet_cache = {}

        service = {}
        with context.session.begin(subtransactions=True):
            LOG.debug('Building service definition entry for %s'
                      % loadbalancer.id)

            # Start with the neutron loadbalancer definition
            (service['loadbalancer'], lb) = self._get_extended_loadbalancer(
                context,
                loadbalancer
            )

            # Get the subnet network associated with the VIP.
            subnet_map = {}
            subnet_id = loadbalancer.vip_subnet_id
            vip_subnet = self._get_subnet_cached(
                context,
                subnet_id
            )
            subnet_map[subnet_id] = vip_subnet

            # Get the network associated with the Loadbalancer.
            network_map = {}
            vip_port = service['loadbalancer']['vip_port']
            network_id = vip_port['network_id']
            service['loadbalancer']['network_id'] = network_id
            network = self._get_network_cached(
                context,
                network_id
            )

            #update和delete时agent为空需要判断
            if agent:
                # Override the segmentation ID and network type for this network
                # if we are running in disconnected service mode
                self._update_network_segment(context, network, agent, vip_port)
            network_map[network_id] = network

            # Check if the tenant can create a loadbalancer on the network.
            if (agent and not self._valid_tenant_ids(network,
                                                     loadbalancer.tenant_id,
                                                     agent)):
                raise sangfor_exc.SangforMismatchedTenants()

            # Get the network VTEPs if the network provider type is
            # either gre or vxlan.
            if 'provider:network_type' in network:
                net_type = network['provider:network_type']
                if net_type == 'vxlan':
                    self._populate_loadbalancer_network_vteps(
                        context,
                        service['loadbalancer'],
                        net_type
                    )

            # Get listeners and pools.
            service['listeners'] = {}
            service['pools'] = {}
            service['l7policies'] = {}
            service['l7rules'] = {}
            service['members'] = {}
            service['healthmonitors'] = {}
            for listener in lb.listeners:
                listener_dict = listener.to_dict(
                    loadbalancer=False,
                    default_pool=False,
                    l7_policies=False,
                )
                listener_dict['l7policies'] = []
                service['listeners'][listener.id] = listener_dict
                for policy in listener.l7_policies:
                    listener_dict['l7policies'].append(policy.id)
                    policy_dict = policy.to_dict(
                        listener=False,
                        redirect_pool=False,
                        rules=False,
                    )
                    policy_dict['rules'] = []
                    service['l7policies'][policy.id] = policy_dict
                    for rule in policy.rules:
                        policy_dict['rules'].append(rule.id)
                        service['l7rules'][rule.id] = rule.to_dict(
                            policy=False,
                        )

            for pool in lb.pools:
                pool_dict = pool.to_dict(
                    healthmonitor=False,
                    sessionpersistence=False,
                    loadbalancer=False,
                    listener=False,
                    listeners=False,
                    l7_policies=False,
                    members=False
                )
                pool_dict['members'] = []
                pool_dict['listeners'] = [ls.id for ls in pool.listeners]

                for member in pool.members:
                    pool_dict['members'].append(member.id)
                    (member_dict, subnet, network) = (
                        self._get_extended_member(context, member)
                    )
                    subnet_map[subnet['id']] = subnet
                    # 这里应该要更新network的层次化绑定信息，但lan ip未生成，可能segment还未生成
                    network_map[network['id']] = network
                    service['members'][member.id] = member_dict

                if pool.healthmonitor is not None:
                    mon = pool.healthmonitor.to_dict(pool=False)
                    mon['pool_id'] = pool.id
                    service['healthmonitors'][mon['id']] = mon

                service['pools'][pool.id] = pool_dict

            service['subnets'] = subnet_map
            service['networks'] = network_map

            return service

    def _update_network_segment(self, context, network, agent, port):
        agent_config = self.deserialize_agent_configurations(
            agent['configurations'])
        segment_data = self.disconnected_service.\
            get_network_segment(context, agent_config,
                                port[portbindings.HOST_ID],
                                port['id'], network)
        if segment_data:
            network['provider:segmentation_id'] = \
                segment_data.get('segmentation_id', None)
            network['provider:network_type'] = \
                segment_data.get('network_type', None)
            network['provider:physical_network'] = \
                segment_data.get('physical_network', None)


    @log_helpers.log_method_call
    def _get_extended_member(self, context, member):
        """Get extended member attributes and member networking."""
        member_dict = member.to_dict(pool=False)
        subnet_id = member.subnet_id
        subnet = self._get_subnet_cached(
            context,
            subnet_id
        )
        network_id = subnet['network_id']
        network = self._get_network_cached(
            context,
            network_id
        )

        member_dict['network_id'] = network_id

        # Use the fixed ip.
        filter = {'fixed_ips': {'subnet_id': [subnet_id],
                                'ip_address': [member.address]}}
        ports = self.plugin.db._core_plugin.get_ports(
            context,
            filter
        )

        # There should be only one.
        if len(ports) == 1:
            member_dict['port'] = ports[0]
            self._populate_member_network(context, member_dict, network)
        else:
            # FIXME(RJB: raise an exception here and let the driver handle
            # the port that is not on the network.
            LOG.error("Unexpected number of ports returned for member: ")
            if not ports:
                LOG.error("No port found")
            else:
                LOG.error("Multiple ports found: %s" % ports)

        return (member_dict, subnet, network)

    @log_helpers.log_method_call
    def _get_extended_loadbalancer(self, context, loadbalancer):
        """Get loadbalancer dictionary and add extended data(e.g. VIP)."""
        lb = self.plugin.db.get_loadbalancer(context, id=loadbalancer.id)
        loadbalancer_dict = lb.to_api_dict()
        vip_port = self.plugin.db._core_plugin.get_port(
            context,
            lb.vip_port_id
        )
        loadbalancer_dict['vip_port'] = vip_port

        return (loadbalancer_dict, lb)

    @log_helpers.log_method_call
    def _get_subnet_cached(self, context, subnet_id):
        """Retrieve subnet from cache if available; otherwise, from Neutron."""
        if subnet_id not in self.subnet_cache:
            subnet = self.plugin.db._core_plugin.get_subnet(
                context,
                subnet_id
            )
            self.subnet_cache[subnet_id] = subnet
        return self.subnet_cache[subnet_id]

    @log_helpers.log_method_call
    def _get_network_cached(self, context, network_id):
        """Retrieve network from cache or from Neutron."""
        if network_id not in self.net_cache:
            network = self.plugin.db._core_plugin.get_network(
                context,
                network_id
            )
            if 'provider:network_type' not in network:
                network['provider:network_type'] = 'undefined'
            if 'provider:segmentation_id' not in network:
                network['provider:segmentation_id'] = 0
            self.net_cache[network_id] = network

        return self.net_cache[network_id]

    @log_helpers.log_method_call
    def _get_listener(self, context, listener_id):
        """Retrieve listener from Neutron db."""
        listener = self.plugin.db.get_listener(
            context,
            listener_id
        )
        return listener.to_api_dict()

    def _populate_member_network(self, context, member, network):
        """Add vtep networking info to pool member and update the network."""
        member['vxlan_vteps'] = []
        if 'provider:network_type' in network:
            net_type = network['provider:network_type']
            if net_type == 'vxlan':
                if 'binding:host_id' in member['port']:
                    host = member['port']['binding:host_id']
                    member['vxlan_vteps'] = self._get_endpoints(
                        context, 'vxlan', host)
        if 'provider:network_type' not in network:
            network['provider:network_type'] = 'undefined'
        if 'provider:segmentation_id' not in network:
            network['provider:segmentation_id'] = 0

    @log_helpers.log_method_call
    def _populate_loadbalancer_network_vteps(
            self,
            context,
            loadbalancer,
            net_type):
        """Put related tunnel endpoints in loadbalancer definiton."""
        loadbalancer['vxlan_vteps'] = self._get_endpoints(context, 'vxlan')

    def _get_endpoints(self, context, net_type, host=None):
        """Get vxlan or gre tunneling endpoints from all agents."""
        endpoints = []

        agents = self.plugin.db._core_plugin.get_agents(context)
        for agent in agents:
            if ('configurations' in agent and (
                    'tunnel_types' in agent['configurations'])):

                if net_type in agent['configurations']['tunnel_types']:
                    if 'tunneling_ip' in agent['configurations']:
                        if not host or (agent['host'] == host):
                            endpoints.append(
                                agent['configurations']['tunneling_ip']
                            )
                    if 'tunneling_ips' in agent['configurations']:
                        for ip_addr in (
                                agent['configurations']['tunneling_ips']):
                            if not host or (agent['host'] == host):
                                endpoints.append(ip_addr)

        return list(set(endpoints))

    def deserialize_agent_configurations(self, configurations):
        """Return a dictionary for the agent configuration."""
        agent_conf = configurations
        if not isinstance(agent_conf, dict):
            try:
                agent_conf = json.loads(configurations)
            except ValueError as ve:
                LOG.error('can not JSON decode %s : %s'
                          % (agent_conf, ve.message))
                agent_conf = {}
        return agent_conf

    @log_helpers.log_method_call
    def _is_common_network(self, network, agent):
        if agent and "configurations" in agent:
            agent_configs = self.deserialize_agent_configurations(
                agent['configurations'])

            common_networks = agent_configs['common_networks']
            """
            common_external_networks = (
                agent_configs['sangfor_common_external_networks'])
            """

            """
            return (network['shared'] or
                    (network['id'] in common_networks) or
                    ('router:external' in network and
                     network['router:external'] and common_external_networks)) 
            return (network['shared'] or
                    (network['id'] in common_networks) or
                    ('router:external' in network and
                     network['router:external'])) 
            """
            return True

    def _valid_tenant_ids(self, network, lb_tenant_id, agent):
        if (network['tenant_id'] == lb_tenant_id):
            return True
        else:
            return self._is_common_network(network, agent)

    @log_helpers.log_method_call
    def _get_ports_on_network(self, context, network_id=None):
        """Get ports for network."""
        if not isinstance(network_id, list):
            network_ids = [network_id]
        filters = {'network_id': network_ids}
        return self.driver.plugin.db._core_plugin.get_ports(
            context,
            filters=filters
        )
